/* planet.c */
/* planet generating program */
/* Copyright 1998 Torben AE. Mogensen */

/* version of August 26 1998 */

/* The program generates planet maps based on recursive spatial subdivision */
/* of a box containing the globe. The output is a colour PPM bitmap. */

/* The colours may optionally be modified according to latitude to move the */
/* icecaps lower closer to the poles, with a corresponding change in land colours. */

/* The Mercator map at magnification 1 is scaled to fit the Width */
/* it uses the full height (it could extend infinitely) */
/* The orthographic projections are scaled so the full view would use the */
/* full Height. Areas outside the globe are coloured black. */
/* Stereographic and gnomic projections use the same scale as orthographic */
/* in the center of the picture, but distorts scale away from the center. */

/* It is assumed that pixels are square */
/* I have included a procedure that will print the maps as ppm */
/* bitmaps (portable pixel map) on standard output. */

/* I have tried to avoid using machine specific features, so it should */
/* be easy to port the program to any machine. Beware, though that due */
/* to different precision on different machines, the same seed numbers */
/* can yield very different planets. */
/* The primitive user interface is a result of portability concerns */

#include <stdio.h>
#include <errno.h>
#include <math.h>

#define BLACK 0
#define WHITE 1
#define BLUE0 2
int BLUE1, LAND0, LAND1;

int BACK = BLACK;

int black_r = 0, black_g = 0, black_b = 0;
int white_r = 255, white_g = 255, white_b = 255;
int blue0_r = 0, blue0_g = 0, blue0_b = 255;
int blue1_r = 0, blue1_g = 128, blue1_b = 255;
int land0_r = 0, land0_g = 255, land0_b = 0;
int land1_r = 128, land1_g = 128, land1_b = 32;

int nocols = 256;

int rtable[256], gtable[256], btable[256];

/* these three values can be changed to change world characteristica */

#define Da0 2.0	 /* initial scale (size of box) */
#define DEG2RAD 0.017453292 /* pi/180 */
#define PI 3.1415926535897

double Dda0 = .13; /* initial altitude variation (suggest 0.13) */
double Ddda = .9;  /* rate of decrease of altitude variation (suggest .9) */

double M  = -.015;  /* initial altitude (slightly below sea level) (suggest -.015) */
int Depth; /* depth of subdivisions */

double t1,t2,t3;
double r1,r2,r3,r4,r5,r6,r7,r8; /* seeds */
double longi,lat,scale;
double vgrid, hgrid;

int latic = 0; /* flag for latitude based colour */

double smooth = 0.0, smooth1 = 1.0; /* amount of smoothing */

int Width = 800, Height = 600;

unsigned char **col;
int **heights;
int cl0[60][10];

int doshade = 0;
int shade;
unsigned char **shades;

double moll_table[] = {0.0, 0.0685055811, 0.1368109534, 0.2047150027,
		       0.2720147303, 0.3385041213, 0.4039727534,
		       0.4682040106, 0.5309726991, 0.5920417499,
		       0.6511575166, 0.7080428038, 0.7623860881,
		       0.8138239166, 0.8619100185, 0.9060553621,
		       0.9453925506, 0.9783738403, 1.0};

double cla, sla, clo, slo;

double rseed, increment = 0.000000001;
int best = 50000;
int weight[11];

int min(x,y)
int x,y;
{ return(x<y ? x : y); }

int max(x,y)
int x,y;
{ return(x<y ? y : x); }

main(ac,av)
int ac;
char **av;
{
  void printppm(), printheights(), print_error();
  void mercator(), peter(), squarep(), mollweide(), sinusoid(), stereo(),
       orthographic(), gnomonic(), azimuth(), conical(), heightfield(), search();
  int i,j, view,c;
  double x,y,z,x1,y1,z1,zz, k1, y2, ymax, ymin;
  double rand2(), log2(), planet1();
  FILE *outfile;
  char filename[256];

  t1 = exp(log(2.)/3.);	   /* dimension of box = (scale,t1*scale,t2*scale) */
  t2 = exp(2.*log(2.)/3.); /* so halving will keep the same relative scale */
  t3 = t2/2.;		   /* like A4 paper (but in 3 dimensions) */

  longi = 0.0;
  lat = 0.0;
  scale = 1.0;
  rseed = 0.123;
  view = 'm';
  vgrid = hgrid = 0.0;
  outfile = stdout;
  
  for (i = 1; i<ac; i++) {
    if (av[i][0] == '-') {
      switch (av[i][1]) {
	case 'V' : sscanf(av[++i],"%lf",&Dda0);
		   break;
	case 'v' : sscanf(av[++i],"%lf",&Ddda);
		   break;
	case 's' : sscanf(av[++i],"%lf",&rseed);
		   break;
	case 'w' : sscanf(av[++i],"%d",&Width);
		   break;
	case 'h' : sscanf(av[++i],"%d",&Height);
		   break;
	case 'm' : sscanf(av[++i],"%lf",&scale);
		   break;
	case 'o' : sscanf(av[++i],"%s",filename);
		   outfile = fopen(filename,"w");
		   if (outfile == NULL) {
		     fprintf(stderr,
			     "Could not open output file %s, error code = %d\n",
			     filename,errno);
		     exit(0);
		   }
		   break;
	case 'l' : sscanf(av[++i],"%lf",&longi);
		   break;
	case 'L' : sscanf(av[++i],"%lf",&lat);
		   break;
	case 'g' : sscanf(av[++i],"%lf",&vgrid);
		   break;
	case 'G' : sscanf(av[++i],"%lf",&hgrid);
		   break;
	case 'c' : latic = 1;
		   break;
        case 'C' : blue1_g = 192; land1_r = 192; land1_g = 192; land1_b = 90;
		   break;
	case 'N' : sscanf(av[++i],"%d",&nocols);
	           if (nocols<4) nocols = 4;
	           if (nocols>255) nocols = 255;
		   break;
	case 'S' : sscanf(av[++i],"%lf",&smooth);
	           if (smooth<0.0) smooth = 0.0;
	           if (smooth>1.0) smooth = 1.0;
		   smooth1 = 1.0-smooth;
	           Dda0 += smooth*Dda0;
		   break;
	case 'B' : doshade = 1;
		   break;
	case 'b' : if (BACK == BLACK) BACK = WHITE; else BACK = BLACK;
		   break;
	case 'i' : sscanf(av[++i],"%lf",&M);
		   break;
	case 'p' : switch (view = av[i][2]) {
		     case 'm' : 
		     case 'p' : 
		     case 'q' : 
		     case 's' :
		     case 'o' :
		     case 'g' :
		     case 'a' :
		     case 'c' :
		     case 'M' : 
		     case 'S' :
		     case 'h' :
		     case 'f' : break;
		     default: fprintf(stderr,"Unknown projection: %s\n",av[i]);
			      print_error();
		   }
		   break;
	default: fprintf(stderr,"Unknown option: %s\n",av[i]);
		 print_error();
      }
    }
    else {
      fprintf(stderr,"Unknown option: %s\n\n",av[i]);
      print_error();
    }
  }

  if (longi>180) longi -= 360;
  longi = longi*DEG2RAD;
  lat = lat*DEG2RAD;

  sla = sin(lat); cla = cos(lat);
  slo = sin(longi); clo = cos(longi);

  if (view == 'h') {
    heights = (int**)calloc(Width,sizeof(int*));
    for (i=0; i<Width; i++)
      heights[i] = (int*)calloc(Height,sizeof(int));
  } else {
    col = (unsigned char**)calloc(Width,sizeof(unsigned char*));
    for (i=0; i<Width; i++)
      col[i] = (unsigned char*)calloc(Height,sizeof(unsigned char));
  }
  if (doshade) {
    shades = (unsigned char**)calloc(Width,sizeof(unsigned char*));
    for (i=0; i<Width; i++)
      shades[i] = (unsigned char*)calloc(Height,sizeof(unsigned char));
  }

  if (view == 'c') {
    if (lat == 0) view = 'm';
        /* Conical approaches mercator when lat -> 0 */
 /*   if (abs(lat) >= PI) view = 's'; */
        /* Conical approaches stereo when lat -> +/- 90 */
  }
  
  if (view == 'f') { /* read map */
    Width = 24; Height = 11;
    for (j = 0; j < Height; j++) {
      y = 7.5*(2.0*j-Height+1);
      y = cos(DEG2RAD*y);
      weight[j] = (int)(100.0*y+0.5);
      for(i = 0; i < Width ; i++) {
	c = getchar();
	switch (c) {
	  case '.': cl0[i][j] = -8;
		    break;
	  case ',': cl0[i][j] = -4;
		    break;
	  case ':': cl0[i][j] = -2;
		    break;
	  case ';': cl0[i][j] = -1;
		    break;
	  case '-': cl0[i][j] = 0;
		    break;
	  case '*': cl0[i][j] = 1;
		    break;
	  case 'o': cl0[i][j] = 2;
		    break;
	  case 'O': cl0[i][j] = 4;
		    break;
	  case '@': cl0[i][j] = 8;
		    break;
	  default: printf("Wrong map symbol: %c\n",c);
	}
      }
      c = getchar(); if (c!='\n') printf("Wrong map format: %c\n",c);
    }
  }

  rtable[BLACK] = 0;
  gtable[BLACK] = 0;
  btable[BLACK] = 0;

  rtable[WHITE] = 255;
  gtable[WHITE] = 255;
  btable[WHITE] = 255;

  BLUE1 = (nocols+2)/2;
  for (i=BLUE0;i<=BLUE1;i++) {
    rtable[i] = (blue0_r*(BLUE1-i)+blue1_r*(i-BLUE0))/(BLUE1-BLUE0);
    gtable[i] = (blue0_g*(BLUE1-i)+blue1_g*(i-BLUE0))/(BLUE1-BLUE0);
    btable[i] = (blue0_b*(BLUE1-i)+blue1_b*(i-BLUE0))/(BLUE1-BLUE0);
  }
  LAND0 = BLUE1+1; LAND1 = nocols-1;
  for (i=LAND0;i<=LAND1;i++) {
    rtable[i] = (land0_r*(LAND1-i)+land1_r*(i-LAND0))/(LAND1-LAND0);
    gtable[i] = (land0_g*(LAND1-i)+land1_g*(i-LAND0))/(LAND1-LAND0);
    btable[i] = (land0_b*(LAND1-i)+land1_b*(i-LAND0))/(LAND1-LAND0);
  }

  Depth = 3*((int)(log2(scale*Height)))+6;


retry:

  r1 = rseed;

  r1 = rand2(r1,r1);
  r2 = rand2(r1,r1);
  r3 = rand2(r1,r2);
  r4 = rand2(r2,r3);
  r5 = rand2(r4,r3);
  r6 = rand2(r4,r5);
  r7 = rand2(r6,r5);
  r8 = rand2(r6,r7);

  switch (view) {

    case 'm': /* Mercator projection */
      mercator();
      break;

    case 'p': /* Peters projection (area preserving cylindrical) */
      peter();
      break;

    case 'q': /* Square projection (equidistant latitudes) */
      squarep();
      break;

    case 'M': /* Mollweide projection (area preserving) */
      mollweide();
      break;

    case 'S': /* Sinusoid projection (area preserving) */
      sinusoid();
      break;

    case 's': /* Stereographic projection */
      stereo();
      break;

    case 'o': /* Orthographic projection */
      orthographic();
      break;

    case 'g': /* Gnomonic projection */
      gnomonic();
      break;

    case 'a': /* Area preserving azimuthal projection */
      azimuth();
      break;

    case 'c': /* Conical projection (conformal) */
      conical();
      break;

    case 'h': /* heightfield */
      heightfield();
      break;

    case 'f': /* Search */
      search();
      rseed += increment;
      goto retry;
  }

  /* plot picture */
  if (view != 'h') printppm(outfile);
  else printheights(outfile);

  exit(0);
}

void mercator()
{
  double x,y,scale1,cos2,theta1, log2();
  int i,j,k;

  y = sin(lat);
  y = (1.0+y)/(1.0-y);
  y = 0.5*log(y);
  k = (int)(0.5*y*Width*scale/PI);
  for (j = 0; j < Height; j++) {
    y = PI*(2.0*(j-k)-Height)/Width/scale;
    y = exp(2.*y);
    y = (y-1.)/(y+1.);
    scale1 = scale*Width/Height/sqrt(1.0-y*y)/PI;
    cos2 = sqrt(1.0-y*y);
    Depth = 3*((int)(log2(scale1*Height)))+3;
    for (i = 0; i < Width ; i++) {
      theta1 = longi-0.5*PI+PI*(2.0*i-Width)/Width/scale;
      col[i][j] = planet0(cos(theta1)*cos2,y,-sin(theta1)*cos2);
      if (doshade) shades[i][j] = shade;
    }
  }
  if (hgrid != 0.0) { /* draw horisontal gridlines */
    for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
    for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
      y = sin(DEG2RAD*theta1);
      y = (1.0+y)/(1.0-y);
      y = 0.5*log(y);
      j = Height/2+(int)(0.5*y*Width*scale/PI)+k;
      if (j>=0 && j<Height) for (i = 0; i < Width ; i++) col[i][j] = BLACK;
    }
  }
  if (vgrid != 0.0) { /* draw vertical gridlines */
    for (theta1 = 0.0; theta1>-360.0; theta1-=vgrid);
    for (theta1 = theta1; theta1<360.0; theta1+=vgrid) {
      i = (int)(0.5*Width*(1.0+scale*(DEG2RAD*theta1-longi)/PI));
      if (i>=0 && i<Width) for (j = 0; j < Height; j++) col[i][j] = BLACK;
    } 
  }
}

void peter()
{
  double x,y,cos2,theta1,scale1, log2();
  int k,i,j,water,land;

  y = 2.0*sin(lat);
  k = (int)(0.5*y*Width*scale/PI);
  water = land = 0;
  for (j = 0; j < Height; j++) {
    y = 0.5*PI*(2.0*(j-k)-Height)/Width/scale;
    if (fabs(y)>1.0)
      for (i = 0; i < Width ; i++) {
	col[i][j] = BACK;
	if (doshade) shades[i][j] = 255;
      }
    else {
      cos2 = sqrt(1.0-y*y);
      if (cos2>0.0) {
	scale1 = scale*Width/Height/cos2/PI;
	Depth = 3*((int)(log2(scale1*Height)))+3;
	for (i = 0; i < Width ; i++) {
	  theta1 = longi-0.5*PI+PI*(2.0*i-Width)/Width/scale;
	  col[i][j] = planet0(cos(theta1)*cos2,y,-sin(theta1)*cos2);
	  if (doshade) shades[i][j] = shade;
	  if (col[i][j] < LAND0) water++; else land++;
        }
      }
    }
  }
  fprintf(stderr,"water percentage: %d\n",100*water/(water+land));
  if (hgrid != 0.0) { /* draw horisontal gridlines */
    for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
    for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
      y = 2.0*sin(DEG2RAD*theta1);
      j = Height/2+(int)(0.5*y*Width*scale/PI)+k;
      if (j>=0 && j<Height) for (i = 0; i < Width ; i++) col[i][j] = BLACK;
    }
  }
  if (vgrid != 0.0) { /* draw vertical gridlines */
    for (theta1 = 0.0; theta1>-360.0; theta1-=vgrid);
    for (theta1 = theta1; theta1<360.0; theta1+=vgrid) {
      i = (int)(0.5*Width*(1.0+scale*(DEG2RAD*theta1-longi)/PI));
      if (i>=0 && i<Width)
	for (j = max(0,Height/2-(int)(Width*scale/PI)+k);
	     j < min(Height,Height/2+(int)(Width*scale/PI)+k); j++)
	  col[i][j] = BLACK;
    }
  }
}

void squarep()
{
  double x,y,scale1,theta1,cos2, log2();
  int k,i,j;

  k = (int)(lat*Width*scale/PI);
  for (j = 0; j < Height; j++) {
    y = (2.0*(j-k)-Height)/Width/scale*PI;
    if (fabs(y)>=0.5*PI) for (i = 0; i < Width ; i++) {
      col[i][j] = BACK;
      if (doshade) shades[i][j] = 255;
    } else {
      cos2 = cos(y);
      if (cos2>0.0) {
	scale1 = scale*Width/Height/cos2/PI;
	Depth = 3*((int)(log2(scale1*Height)))+3;
	for (i = 0; i < Width ; i++) {
	  theta1 = longi-0.5*PI+PI*(2.0*i-Width)/Width/scale;
	  col[i][j] = planet0(cos(theta1)*cos2,sin(y),-sin(theta1)*cos2);
	  if (doshade) shades[i][j] = shade;
        }
      }
    }
  }
  if (hgrid != 0.0) { /* draw horisontal gridlines */
    for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
    for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
      y = DEG2RAD*theta1;
      j = Height/2+(int)(0.5*y*Width*scale/PI)+k;
      if (j>=0 && j<Height) for (i = 0; i < Width ; i++) col[i][j] = BLACK;
    }
  }
  if (vgrid != 0.0) { /* draw vertical gridlines */
    for (theta1 = 0.0; theta1>-360.0; theta1-=vgrid);
    for (theta1 = theta1; theta1<360.0; theta1+=vgrid) {
      i = (int)(0.5*Width*(1.0+scale*(DEG2RAD*theta1-longi)/PI));
      if (i>=0 && i<Width)
	for (j = max(0,Height/2-(int)(0.25*PI*Width*scale/PI)+k);
	     j < min(Height,Height/2+(int)(0.25*PI*Width*scale/PI)+k); j++)
	  col[i][j] = BLACK;
    }
  }
}

void mollweide()
{
  double x,y,y1,zz,scale1,cos2,theta1,theta2, log2();
  int i,j,i1,k;

  for (j = 0; j < Height; j++) {
    y1 = 2*(2.0*j-Height)/Width/scale;
    if (fabs(y1)>=1.0) for (i = 0; i < Width ; i++) {
      col[i][j] = BACK;
      if (doshade) shades[i][j] = 255;
    } else {
      zz = sqrt(1.0-y1*y1);
      y = 2.0/PI*(y1*zz+asin(y1));
      cos2 = sqrt(1.0-y*y);
      if (cos2>0.0) {
	scale1 = scale*Width/Height/cos2/PI;
	Depth = 3*((int)(log2(scale1*Height)))+3;
	for (i = 0; i < Width ; i++) {
	  theta1 = PI/zz*(2.0*i-Width)/Width/scale;
	  if (fabs(theta1)>PI) {
	    col[i][j] = BACK;
	    if (doshade) shades[i][j] = 255;
	  } else {
	    theta1 += longi-0.5*PI;
	    col[i][j] = planet0(cos(theta1)*cos2,y,-sin(theta1)*cos2);
	    if (doshade) shades[i][j] = shade;
	  }
        }
      }
    }
  }
  if (hgrid != 0.0) { /* draw horisontal gridlines */
    for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
    for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
      theta2 = fabs(theta1);
      x = floor(theta2/5.0); y = theta2/5.0-x;
      y = (1.0-y)*moll_table[(int)x]+y*moll_table[(int)x+1];
      if (theta1<0.0) y = -y;
      j = Height/2+(int)(0.25*y*Width*scale);
      if (j>=0 && j< Height) {
	for (i = max(0,Width/2-(int)(0.5*Width*scale*sqrt(1.0-y*y)));
	     i < min(Width,Width/2+(int)(0.5*Width*scale*sqrt(1.0-y*y))); i++)
	  col[i][j] = BLACK;
      }
    }
  }
  if (vgrid != 0.0) { /* draw vertical gridlines */
    for (theta1 = 0.0; theta1>-360.0; theta1-=vgrid);
    for (theta1 = theta1; theta1<360.0; theta1+=vgrid) {
      if (DEG2RAD*theta1-longi+0.5*PI>-PI &&
	  DEG2RAD*theta1-longi+0.5*PI<=PI) {
	x = 0.5*(DEG2RAD*theta1-longi+0.5*PI)*Width*scale/PI;
	j = max(0,Height/2-(int)(0.25*Width*scale));
	y = 2*(2.0*j-Height)/Width/scale;
	i = Width/2+x*sqrt(1.0-y*y);
	for (; j <= min(Height,Height/2+(int)(0.25*Width*scale)); j++) {
	  y1 = 2*(2.0*j-Height)/Width/scale;
	  if (fabs(y1)<=1.0) {
	    i1 = Width/2+x*sqrt(1.0-y1*y1);
	    if (i1>0 && i1<=Width) col[i1][j] = BLACK;
	  }
	  if (fabs(y)<=1.0) {
	    if (i<i1) {
	      for (k=i+1; k<i1; k++)
		if (k>0 && k<=Width) col[k][j] = BLACK;
	    }
	    else if (i>i1) {
	      for (k=i-1; k>i1; k--)
		if (k>0 && k<=Width) col[k][j] = BLACK;
	    }
	  }
	  y = y1;
	  i = i1;
        }
      }
    }
  }
}

void sinusoid()
{
  double x,y,theta1,theta2,cos2,l1,i1,scale1, log2();
  int k,i,j,l,c;

  k = (int)(lat*Width*scale/PI);
  for (j = 0; j < Height; j++) {
    y = (2.0*(j-k)-Height)/Width/scale*PI;
    if (fabs(y)>=0.5*PI) for (i = 0; i < Width ; i++) {
      col[i][j] = BACK;
      if (doshade) shades[i][j] = 255;
    } else {
      cos2 = cos(y);
      if (cos2>0.0) {
	scale1 = scale*Width/Height/cos2/PI;
	Depth = 3*((int)(log2(scale1*Height)))+3;
	for (i = 0; i<Width; i++) {
	  l = i*12/Width;
	  l1 = l*Width/12.0;
	  i1 = i-l1;
	  theta2 = longi-0.5*PI+PI*(2.0*l1-Width)/Width/scale;
	  theta1 = (PI*(2.0*i1-Width/12)/Width/scale)/cos2;
	  if (fabs(theta1)>PI/12.0) {
	    col[i][j] = BACK;
	    if (doshade) shades[i][j] = 255;
	  } else {
	    col[i][j] =
		planet0(cos(theta1+theta2)*cos2,sin(y),
			-sin(theta1+theta2)*cos2);
	    if (doshade) shades[i][j] = shade;
	  }
        }
      }
    }
  }
  if (hgrid != 0.0) { /* draw horisontal gridlines */
    for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
    for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
      y = DEG2RAD*theta1;
      cos2 = cos(y);
      j = Height/2+(int)(0.5*y*Width*scale/PI)+k;
      if (j>=0 && j<Height)
	for (i = 0; i < Width ; i++) {
	  l = i*12/Width;
	  l1 = l*Width/12.0;
	  i1 = i-l1;
	  theta2 = (PI*(2.0*i1-Width/12)/Width/scale)/cos2;
	  if (fabs(theta2)<=PI/12.0) col[i][j] = BLACK;
        }
    }
  }
  if (vgrid != 0.0) { /* draw vertical gridlines */
    for (theta1 = 0.0; theta1>-360.0; theta1-=vgrid);
    for (theta1 = theta1; theta1<360.0; theta1+=vgrid) {
      i = (int)(0.5*Width*(1.0+scale*(DEG2RAD*theta1-longi)/PI));
      if (i>=0 && i<Width)
	for (j = max(0,Height/2-(int)(0.25*PI*Width*scale/PI)+k);
	   j < min(Height,Height/2+(int)(0.25*PI*Width*scale/PI)+k); j++) {
	    y = (2.0*(j-k)-Height)/Width/scale*PI;
	    cos2 = cos(y);
	    l = i*12/Width;
	    l1 = l*Width/12.0+Width/24.0;
	    i1 = i-l1;
	    c = l1+i1*cos2;
	    if (c>=0 && c<Width) col[c][j] = BLACK;
	  }
    }
  }
}

void stereo()
{
  double x,y,ymin,ymax,z,zz,x1,y1,z1,theta1,theta2;
  int i,j;

  ymin = 2.0;
  ymax = -2.0;
  for (j = 0; j < Height; j++) {
    for (i = 0; i < Width ; i++) {
      x = (2.0*i-Width)/Height/scale;
      y = (2.0*j-Height)/Height/scale;
      z = x*x+y*y;
      zz = 0.25*(4.0+z);
      x = x/zz;
      y = y/zz;
      z = (1.0-0.25*z)/zz;
      x1 = clo*x+slo*sla*y+slo*cla*z;
      y1 = cla*y-sla*z;
      z1 = -slo*x+clo*sla*y+clo*cla*z;
      if (y1 < ymin) ymin = y1;
      if (y1 > ymax) ymax = y1;
      col[i][j] = planet0(x1,y1,z1);
      if (doshade) shades[i][j] = shade;
    }
  }
  if (hgrid != 0.0) { /* draw horisontal gridlines */
    for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
    for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
      y = sin(DEG2RAD*theta1);
      if (ymin <= y && y <= ymax) {
	zz = sqrt(1-y*y);
	for (theta2=-PI; theta2<PI; theta2+=0.5/Width/scale) {
	  x = sin(theta2)*zz;
	  z = cos(theta2)*zz;
	  x1 = clo*x+slo*z;
	  y1 = slo*sla*x+cla*y-clo*sla*z;
	  z1 = -slo*cla*x+sla*y+clo*cla*z;
	  if (fabs(z1)<1.0){
	    i = 0.5*(Height*scale*2.0*x1*(1+z1)/(1.0-z1*z1)+Width);
	    j = 0.5*(Height*scale*2.0*y1*(1+z1)/(1.0-z1*z1)+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
  if (vgrid != 0.0) { /* draw vertical gridlines */
    for (theta2=-PI; theta2<PI; theta2+=0.5/Width/scale) {
      y = sin(theta2);
      if (ymin <= y && y <= ymax) {
	for (theta1 = 0.0; theta1<360.0; theta1+=vgrid) {
	  x = sin(DEG2RAD*theta1)*cos(theta2);
	  z = cos(DEG2RAD*theta1)*cos(theta2);
	  x1 = clo*x+slo*z;
	  y1 = slo*sla*x+cla*y-clo*sla*z;
	  z1 = -slo*cla*x+sla*y+clo*cla*z;
	  if (fabs(z1)<1.0){
	    i = 0.5*(Height*scale*2.0*x1*(1+z1)/(1-z1*z1)+Width);
	    j = 0.5*(Height*scale*2.0*y1*(1+z1)/(1-z1*z1)+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
}

void orthographic()
{
  double x,y,z,x1,y1,z1,ymin,ymax,theta1,theta2,zz;
  int i,j;

  ymin = 2.0;
  ymax = -2.0;
  for (j = 0; j < Height; j++) {
    for (i = 0; i < Width ; i++) {
      x = (2.0*i-Width)/Height/scale;
      y = (2.0*j-Height)/Height/scale;
      if (x*x+y*y>1.0) {
	col[i][j] = BACK;
	if (doshade) shades[i][j] = 255;
      } else {
	z = sqrt(1.0-x*x-y*y);
	x1 = clo*x+slo*sla*y+slo*cla*z;
	y1 = cla*y-sla*z;
	z1 = -slo*x+clo*sla*y+clo*cla*z;
	if (y1 < ymin) ymin = y1;
	if (y1 > ymax) ymax = y1;
	col[i][j] = planet0(x1,y1,z1);
	if (doshade) shades[i][j] = shade;
      }
    }
  }
  if (hgrid != 0.0) { /* draw horisontal gridlines */
    for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
    for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
      y = sin(DEG2RAD*theta1);
      if (ymin <= y && y <= ymax) {
	zz = sqrt(1-y*y);
	for (theta2=-PI; theta2<PI; theta2+=0.5/Width/scale) {
	  x = sin(theta2)*zz;
	  z = cos(theta2)*zz;
	  x1 = clo*x+slo*z;
	  y1 = slo*sla*x+cla*y-clo*sla*z;
	  z1 = -slo*cla*x+sla*y+clo*cla*z;
	  if (0.0>=z1){
	    i = 0.5*(Height*scale*x1+Width);
	    j = 0.5*(Height*scale*y1+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
  if (vgrid != 0.0) { /* draw vertical gridlines */
    for (theta2=-PI; theta2<PI; theta2+=0.5/Width/scale) {
      y = sin(theta2);
      if (ymin <= y && y <= ymax) {
	for (theta1 = 0.0; theta1<360.0; theta1+=vgrid) {
	  x = sin(DEG2RAD*theta1)*cos(theta2);
	  z = cos(DEG2RAD*theta1)*cos(theta2);
	  x1 = clo*x+slo*z;
	  y1 = slo*sla*x+cla*y-clo*sla*z;
	  z1 = -slo*cla*x+sla*y+clo*cla*z;
	  if (0.0>=z1){
	    i = 0.5*(Height*scale*x1+Width);
	    j = 0.5*(Height*scale*y1+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
}

void gnomonic()
{
  double x,y,z,x1,y1,z1,zz,theta1,theta2,ymin,ymax;
  int i,j;

  ymin = 2.0;
  ymax = -2.0;
  for (j = 0; j < Height; j++) {
    for (i = 0; i < Width ; i++) {
      x = (2.0*i-Width)/Height/scale;
      y = (2.0*j-Height)/Height/scale;
      zz = sqrt(1.0/(1.0+x*x+y*y));
      x = x*zz;
      y = y*zz;
      z = sqrt(1.0-x*x-y*y);
      x1 = clo*x+slo*sla*y+slo*cla*z;
      y1 = cla*y-sla*z;
      z1 = -slo*x+clo*sla*y+clo*cla*z;
      if (y1 < ymin) ymin = y1;
      if (y1 > ymax) ymax = y1;
      col[i][j] = planet0(x1,y1,z1);
      if (doshade) shades[i][j] = shade;
    }
  }
  if (hgrid != 0.0) { /* draw horisontal gridlines */
    for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
    for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
      y = sin(DEG2RAD*theta1);
      if (ymin <= y && y <= ymax) {
	zz = sqrt(1-y*y);
	for (theta2=-PI; theta2<PI; theta2+=0.5/Width/scale) {
	  x = sin(theta2)*zz;
	  z = cos(theta2)*zz;
	  x1 = clo*x-slo*z;
	  y1 = slo*sla*x+cla*y+clo*sla*z;
	  z1 = slo*cla*x-sla*y+clo*cla*z;
	  if (z1!=0.0){
	    i = 0.5*(Height*scale*x1/z1+Width);
	    j = 0.5*(Height*scale*y1/z1+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
  if (vgrid != 0.0) { /* draw vertical gridlines */
    for (theta2=-PI; theta2<PI; theta2+=0.5/Width/scale) {
      y = sin(theta2);
      if (ymin <= y && y <= ymax) {
	for (theta1 = 0.0; theta1<360.0; theta1+=vgrid) {
	  x = sin(DEG2RAD*theta1)*cos(theta2);
	  z = cos(DEG2RAD*theta1)*cos(theta2);
	  x1 = clo*x-slo*z;
	  y1 = slo*sla*x+cla*y+clo*sla*z;
	  z1 = slo*cla*x-sla*y+clo*cla*z;
	  if (z1!=0.0){
	    i = 0.5*(Height*scale*x1/z1+Width);
	    j = 0.5*(Height*scale*y1/z1+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
}

void azimuth()
{
  double x,y,z,x1,y1,z1,zz,theta1,theta2,ymin,ymax;
  int i,j;

  ymin = 2.0;
  ymax = -2.0;
  for (j = 0; j < Height; j++) {
    for (i = 0; i < Width ; i++) {
      x = (2.0*i-Width)/Height/scale;
      y = (2.0*j-Height)/Height/scale;
      zz = x*x+y*y;
      z = 1.0-0.5*zz;
      if (z<-1.0) {
	col[i][j] = BACK;
	if (doshade) shades[i][j] = 255;
      } else {
	zz = sqrt(1.0-0.25*zz);
	x = x*zz;
	y = y*zz;
	x1 = clo*x+slo*sla*y+slo*cla*z;
	y1 = cla*y-sla*z;
	z1 = -slo*x+clo*sla*y+clo*cla*z;
	if (y1 < ymin) ymin = y1;
	if (y1 > ymax) ymax = y1;
	col[i][j] = planet0(x1,y1,z1);
	if (doshade) shades[i][j] = shade;
      }
    }
  }
  if (hgrid != 0.0) { /* draw horisontal gridlines */
    for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
    for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
      y = sin(DEG2RAD*theta1);
      if (ymin <= y && y <= ymax) {
	zz = sqrt(1-y*y);
	for (theta2=-PI; theta2<PI; theta2+=0.5/Width/scale) {
	  x = sin(theta2)*zz;
	  z = cos(theta2)*zz;
	  x1 = clo*x-slo*z;
	  y1 = slo*sla*x+cla*y+clo*sla*z;
	  z1 = slo*cla*x-sla*y+clo*cla*z;
	  if (z1!=-1.0){
	    i = 0.5*(Height*scale*x1/sqrt(0.5+0.5*z1)+Width);
	    j = 0.5*(Height*scale*y1/sqrt(0.5+0.5*z1)+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
  if (vgrid != 0.0) { /* draw vertical gridlines */
    for (theta2=-PI; theta2<PI; theta2+=0.5/Width/scale) {
      y = sin(theta2);
      if (ymin <= y && y <= ymax) {
	for (theta1 = 0.0; theta1<360.0; theta1+=vgrid) {
	  x = sin(DEG2RAD*theta1)*cos(theta2);
	  z = cos(DEG2RAD*theta1)*cos(theta2);
	  x1 = clo*x-slo*z;
	  y1 = slo*sla*x+cla*y+clo*sla*z;
	  z1 = slo*cla*x-sla*y+clo*cla*z;
	  if (z1!=-1.0){
	    i = 0.5*(Height*scale*x1/sqrt(0.5+0.5*z1)+Width);
	    j = 0.5*(Height*scale*y1/sqrt(0.5+0.5*z1)+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
}

void conical()
{
  double k1,c,y2,x,y,z,zz,x1,y1,z1,theta1,theta2,ymin,ymax,cos2;
  int i,j;

  ymin = 2.0;
  ymax = -2.0;
  if (lat>0) {
    k1 = 1.0/sin(lat);
    c = k1*k1;
    y2 = sqrt(c*(1.0-sin(lat/k1))/(1.0+sin(lat/k1)));
    for (j = 0; j < Height; j++) {
      for (i = 0; i < Width ; i++) {
	x = (2.0*i-Width)/Height/scale;
	y = (2.0*j-Height)/Height/scale+y2;
	zz = x*x+y*y;
	if (zz==0.0) theta1 = 0.0; else theta1 = k1*atan2(x,y);
	if (theta1<-PI || theta1>PI) {
	  col[i][j] = BACK;
	  if (doshade) shades[i][j] = 255;
	} else {
	  theta1 += longi-0.5*PI; /* theta1 is longitude */
	  theta2 = k1*asin((zz-c)/(zz+c));
	  /* theta2 is latitude */
	  if (theta2 > 0.5*PI || theta2 < -0.5*PI) {
	    col[i][j] = BACK;
	    if (doshade) shades[i][j] = 255;
	  } else {
	    cos2 = cos(theta2);
	    y = sin(theta2);
	    if (y < ymin) ymin = y;
	    if (y > ymax) ymax = y;
	    col[i][j] = planet0(cos(theta1)*cos2,y,-sin(theta1)*cos2);
	    if (doshade) shades[i][j] = shade;
	  }
        }
      }
    }
    if (hgrid != 0.0) { /* draw horisontal gridlines */
      for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
      for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
	y = sin(DEG2RAD*theta1);
	if (ymin <= y && y <= ymax) {
	  zz = sqrt(c*(1.0+sin(DEG2RAD*theta1/k1))
		    /(1.0-sin(DEG2RAD*theta1/k1)));
	  for (theta2=-PI+longi; theta2<PI+longi; theta2+=0.5/Width/scale) {
	    z1 = theta2-longi;
	    x1 = zz*sin(z1/k1);
	    y1 = zz*cos(z1/k1);
	    i = 0.5*(Height*scale*x1+Width);
	    j = 0.5*(Height*scale*(y1-y2)+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
    if (vgrid != 0.0) { /* draw vertical gridlines */
      for (theta1=-0.5*PI; theta1<0.5*PI; theta1+=0.5/Width/scale) {
	y = sin(theta1);
	if (ymin <= y && y <= ymax) {
	  zz = sqrt(c*(1.0+sin(theta1/k1))
		    /(1.0-sin(theta1/k1)));
	  for (theta2 = 0.0; theta2>-180.0+longi/DEG2RAD; theta2-=vgrid);
	  for (theta2 = theta2; theta2<180.0+longi/DEG2RAD; theta2+=vgrid) {
	    z1 = DEG2RAD*theta2-longi;
	    x1 = zz*sin(z1/k1);
	    y1 = zz*cos(z1/k1);
	    i = 0.5*(Height*scale*x1+Width);
	    j = 0.5*(Height*scale*(y1-y2)+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
  else {
    k1 = 1.0/sin(lat);
    c = k1*k1;
    y2 = sqrt(c*(1.0-sin(lat/k1))/(1.0+sin(lat/k1)));
    for (j = 0; j < Height; j++) {
      for (i = 0; i < Width ; i++) {
	x = (2.0*i-Width)/Height/scale;
	y = (2.0*j-Height)/Height/scale-y2;
	zz = x*x+y*y;
	if (zz==0.0) theta1 = 0.0; else theta1 = -k1*atan2(x,-y);
	if (theta1<-PI || theta1>PI) {
	  col[i][j] = BACK;
	  if (doshade) shades[i][j] = 255;
	} else {
	  theta1 += longi-0.5*PI; /* theta1 is longitude */
	  theta2 = k1*asin((zz-c)/(zz+c));
	  /* theta2 is latitude */
	  if (theta2 > 0.5*PI || theta2 < -0.5*PI) {
	    col[i][j] = BACK;
	    if (doshade) shades[i][j] = 255;
	  } else {
	    cos2 = cos(theta2);
	    y = sin(theta2);
	    if (y < ymin) ymin = y;
	    if (y > ymax) ymax = y;
	    col[i][j] = planet0(cos(theta1)*cos2,y,-sin(theta1)*cos2);
	    if (doshade) shades[i][j] = shade;
	  }
        }
      }
    }
    if (hgrid != 0.0) { /* draw horisontal gridlines */
      for (theta1 = 0.0; theta1>-90.0; theta1-=hgrid);
      for (theta1 = theta1; theta1<90.0; theta1+=hgrid) {
	y = sin(DEG2RAD*theta1);
	if (ymin <= y && y <= ymax) {
	  zz = sqrt(c*(1.0+sin(DEG2RAD*theta1/k1))
		    /(1.0-sin(DEG2RAD*theta1/k1)));
	  for (theta2=-PI+longi; theta2<PI+longi; theta2+=0.5/Width/scale) {
	    z1 = theta2-longi;
	    x1 = -zz*sin(z1/k1);
	    y1 = -zz*cos(z1/k1);
	    i = 0.5*(Height*scale*x1+Width);
	    j = 0.5*(Height*scale*(y1+y2)+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
    if (vgrid != 0.0) { /* draw vertical gridlines */
      for (theta1=-0.5*PI; theta1<0.5*PI; theta1+=0.5/Width/scale) {
	y = sin(theta1);
	if (ymin <= y && y <= ymax) {
	  zz = sqrt(c*(1.0+sin(theta1/k1))
		    /(1.0-sin(theta1/k1)));
	  for (theta2 = 0.0; theta2>-180.0+longi/DEG2RAD; theta2-=vgrid);
	  for (theta2 = theta2; theta2<180.0+longi/DEG2RAD; theta2+=vgrid) {
	    z1 = DEG2RAD*theta2-longi;
	    x1 = -zz*sin(z1/k1);
	    y1 = -zz*cos(z1/k1);
	    i = 0.5*(Height*scale*x1+Width);
	    j = 0.5*(Height*scale*(y1+y2)+Height);
	    if (0<=i && i<Width && 0<=j && j<Height) col[i][j] = BLACK;
	  }
        }
      }
    }
  }
}


void heightfield()
{
  double x,y,z,x1,y1,z1, planet1();
  int i,j;

  for (j = 0; j < Height; j++) {
    for (i = 0; i < Width ; i++) {
      x = (2.0*i-Width)/Height/scale;
      y = (2.0*j-Height)/Height/scale;
      if (x*x+y*y>1.0) heights[i][j] = 0;
      else {
	z = sqrt(1.0-x*x-y*y);
	x1 = clo*x+slo*sla*y+slo*cla*z;
	y1 = cla*y-sla*z;
	z1 = -slo*x+clo*sla*y+clo*cla*z;
	heights[i][j] = 10000000*planet1(x1,y1,z1);
      }
    }
  }
}


void search()
{
  double y,cos2,theta1,scale1, planet1(), log2();
  int i,j,k,l,c,errcount, errcount1;

  for (j = 0; j < Height; j++) {
    y = 7.5*(2.0*j-Height+1);
    y = sin(DEG2RAD*y);
    scale1 = Width/Height/sqrt(1.0-y*y)/PI;
    cos2 = sqrt(1.0-y*y);
    Depth = 3*((int)(log2(scale1*Height)))+6;
    for (i = 0; i < Width ; i++) {
      theta1 = -0.5*PI+PI*(2.0*i-Width)/Width;
      c = 128+1000*planet1(cos(theta1)*cos2,y,-sin(theta1)*cos2);
      if (c<0) c = 0;
      if (c>255) c = 255;
      col[i][j] = c;
    }
  }
  for (k=0; k<Width; k++) {
    for (l=-20; l<=20; l+=2) {
      errcount = 0;
      for (j = 0; j < Height; j++) {
	errcount1 = 0;
	for(i = 0; i < Width ; i++) {
	  if (cl0[i][j]<0 && col[(i+k)%Width][j] > 128-l)
	    errcount1-=cl0[i][j];
	  if (cl0[i][j]>0 && col[(i+k)%Width][j] <= 128-l)
	    errcount1+=cl0[i][j];
        }
	errcount += weight[j]*errcount1;
      }

      if (errcount < best) {
	printf("Errors: %d, seed: %.16f, rotate = %.1f, initial altitude = %.3f\n",
	       errcount,rseed,(360.0*k)/Width,M+l/1000.0);
	best = errcount;
	for (j = 0; j < Height; j++) {
	  for(i = 0; i < Width ; i++)
	    if (col[(i+k)%Width][j] <= 128-l) putchar('.');
	    else putchar('O');
	  putchar('\n');
        }
	fflush(stdout);
      }
    }
  }
}

int planet0(x,y,z)
double x,y,z;
{
  double alt, planet1();
  int colour;

  alt = planet1(x,y,z);

      /* calculate colour */
  if (alt <=0.) { /* if below sea level then */
    if (latic && y*y+alt >= 0.98)
      colour = WHITE;	 /* white if close to poles */
    else {
      colour = BLUE1+(int)((BLUE1-BLUE0+1)*(10*alt));	/* blue scale otherwise */
      if (colour<BLUE0) colour = BLUE0;
    }
  }
  else {
    if (latic) alt += 0.10204*y*y;  /* altitude adjusted with latitude */
    if (alt >= 0.1) /* if high then */
      colour = WHITE;
    else {
      colour = LAND0+(int)((LAND1-LAND0+1)*(10*alt));
	    /* else green to brown scale */
      if (colour>LAND1) colour = LAND1;
    }
  }
  return(colour);
}

double ta,tb,tc,td,te,tf,tg,th;
double tas,tbs,tcs,tds,tes,tfs,tgs,ths;
double tax,tay,taz, tscale, tdalt;

double planet(a,b,c,d,e,f,g,h, as,bs,cs,ds,es,fs,gs,hs,
       ax,ay,az, x,y,z, level, scale, dalt)
double a,b,c,d,e,f,g,h;		/* altitudes of the eight corners */
double as,bs,cs,ds,es,fs,gs,hs; /* seeds of the eight corners */
double ax,ay,az;		/* bottom left closest corner */
double x,y,z, scale, dalt;	/* goal point, scale and variance */
int level;
{
  double ae,bf,cg,dh,aes,bfs,cgs,dhs,tmp,rand2();
  double x1,y1,z1,l1,l2;
  double a0,b0,c0,d0,e0,f0,g0,h0;
  double ddd, ddd1;

  a0=a; b0=b; c0=c; d0=d; e0=e; f0=f; g0=g; h0=h;
  ddd = Ddda*Ddda*Ddda; ddd1 = 1.0/(1.0-ddd);

  while (level>0) {
    if (level == 12) { /* save information for shortcut */
      ta=a; tb=b; tc=c; td=d; te=e; tf=f; tg=g; th=h;
      tas=as; tbs=bs; tcs=cs; tds=ds; tes=es; tfs=fs; tgs=gs; ths=hs;
      tax=ax; tay=ay; taz=az; tscale=scale; tdalt=dalt;
      a0=a; b0=b; c0=c; d0=d; e0=e; f0=f; g0=g; h0=h;
    }

    /* subdivide over longest axis */
    aes = rand2(as,es); bfs = rand2(bs,fs); /* find seeds for new points */
    cgs = rand2(cs,gs); dhs = rand2(ds,hs);
    /* find altitudes as mean+variation, using smoothing */
    ae = (a+e+(smooth1*dalt+smooth*fabs(a-e))*aes)/2.;
    bf = (b+f+(smooth1*dalt+smooth*fabs(b-f))*bfs)/2.;
    cg = (c+g+(smooth1*dalt+smooth*fabs(c-g))*cgs)/2.;
    dh = (d+h+(smooth1*dalt+smooth*fabs(d-h))*dhs)/2.;
          /* find which halfbox goal point is in and continue in that */
    if (az+scale*t3 >= z) {
      h=c; g=cg; f=dh; e=d; d=b; c=bf; b=ae;
      hs=cs; gs=cgs; fs=dhs; es=ds; ds=bs; cs=bfs; bs=aes;
      tmp = ax; ax=az; az=ay; ay=tmp;
    }
    else {
      a=ae; b=e; c=f; d=bf; e=dh; f=h; h=cg;
      as=aes; bs=es; cs=fs; ds=bfs; es=dhs; fs=hs; hs=cgs;
      tmp = ax; ax=az+scale*t3; az=ay; ay=tmp;
    }
    tmp = x; x=z; z=y; y=tmp;
    scale=scale*t3; dalt=dalt*Ddda;

    /* subdivide over longest axis */
    aes = rand2(as,es); bfs = rand2(bs,fs); /* find seeds for new points */
    cgs = rand2(cs,gs); dhs = rand2(ds,hs);
    /* find altitudes as mean+variation, using smoothing */
    ae = (a+e+(smooth1*dalt+smooth*fabs(a-e))*aes)/2.;
    bf = (b+f+(smooth1*dalt+smooth*fabs(b-f))*bfs)/2.;
    cg = (c+g+(smooth1*dalt+smooth*fabs(c-g))*cgs)/2.;
    dh = (d+h+(smooth1*dalt+smooth*fabs(d-h))*dhs)/2.;
          /* find which halfbox goal point is in and continue in that */
    if (az+scale*t3 >= z) {
      h=c; g=cg; f=dh; e=d; d=b; c=bf; b=ae;
      hs=cs; gs=cgs; fs=dhs; es=ds; ds=bs; cs=bfs; bs=aes;
      tmp = ax; ax=az; az=ay; ay=tmp;
    }
    else {
      a=ae; b=e; c=f; d=bf; e=dh; f=h; h=cg;
      as=aes; bs=es; cs=fs; ds=bfs; es=dhs; fs=hs; hs=cgs;
      tmp = ax; ax=az+scale*t3; az=ay; ay=tmp;
    }
    tmp = x; x=z; z=y; y=tmp;
    scale=scale*t3; dalt=dalt*Ddda;

    /* subdivide over longest axis */
    aes = rand2(as,es); bfs = rand2(bs,fs); /* find seeds for new points */
    cgs = rand2(cs,gs); dhs = rand2(ds,hs);
    /* find altitudes as mean+variation, using smoothing */
    ae = (a+e+(smooth1*dalt+smooth*fabs(a-e))*aes)/2.;
    bf = (b+f+(smooth1*dalt+smooth*fabs(b-f))*bfs)/2.;
    cg = (c+g+(smooth1*dalt+smooth*fabs(c-g))*cgs)/2.;
    dh = (d+h+(smooth1*dalt+smooth*fabs(d-h))*dhs)/2.;
          /* find which halfbox goal point is in and continue in that */
    if (az+scale*t3 >= z) {
      h=c; g=cg; f=dh; e=d; d=b; c=bf; b=ae;
      hs=cs; gs=cgs; fs=dhs; es=ds; ds=bs; cs=bfs; bs=aes;
      tmp = ax; ax=az; az=ay; ay=tmp;
    }
    else {
      a=ae; b=e; c=f; d=bf; e=dh; f=h; h=cg;
      as=aes; bs=es; cs=fs; ds=bfs; es=dhs; fs=hs; hs=cgs;
      tmp = ax; ax=az+scale*t3; az=ay; ay=tmp;
    }
    tmp = x; x=z; z=y; y=tmp;
    scale=scale*t3; dalt=dalt*Ddda;

    a0=ddd*a0+a; b0=ddd*b0+b; c0=ddd*c0+c; d0=ddd*d0+d;
    e0=ddd*e0+e; f0=ddd*f0+f; g0=ddd*g0+g; h0=ddd*h0+h;

    level = level-3;
  }
  if (doshade) {
    x1 = x+(a0-b0)/dalt/ddd1;
    y1 = y+(a0-d0)/t1/dalt/ddd1;
    z1 = z+(a0-e0)/t2/dalt/ddd1;
    l1 = sqrt(x1*x1+y1*y1+z1*z1);
    l2 = sqrt(x*x+z*z);
    if (l2==0) shade = 128;
    else shade = (int)((-z*x1+x*z1)/l1/l2*128.0+128.0);
    if (shade<10) shade = 10;
    if (shade>255) shade = 255;
  }
  else return((a+b+c+d+e+f+g+h)/8);
}


double planet1(x,y,z)
double x,y,z;
{
  if (tax<x && tay<y && taz<z &&
      x<tax+tscale && y<tay+tscale*t1 && z<taz+tscale*t2) /* take shortcut */
    return(planet(ta,tb,tc,td,te,tf,tg,th,
		  tas,tbs,tcs,tds,tes,tfs,tgs,ths,
		  tax,tay,taz, x,y,z, 12, tscale, tdalt));
  else return(planet(M,M,M,M,M,M,M,M,
		     /* initial altitude is M on all corners of box */
		     r1,r2,r3,r4,r5,r6,r7,r8,
		     /* same seed set is used in every call */
		     -1.,-t1,-t2,
		     /* bottom left closest corner: (-1,-2^(1/3),-2^(2/3)) */
		     x,y,z,
		     /* coordinates of point we want colour of */
		     Depth,Da0,Dda0));
		     /* subdivision depth, initial scale and variance */
}


double rand2(p,q) /* random number generator taking two seeds */
double p,q;	  /* rand2(p,q) = rand2(q,p) is important     */
{
  double r;
  r = (p+3.14159265)*(q+3.14159265);
  return(2.*(r-(int)r)-1.);
}

void printppm(outfile) /* prints picture in PPM (portable pixel map) format */
FILE *outfile;
{
  int i,j,c;

  fprintf(outfile,"P6\n");
  fprintf(outfile,"#fractal planet image\n");
  fprintf(outfile,"%d %d 255\n",Width,Height);
 
  if (doshade) {
    for (j=0; j<Height; j++)
      for (i=0; i<Width; i++) {
	c = shades[i][j]*rtable[col[i][j]]/150;
	if (c>255) c=255;
	putc(c,outfile);
	c = shades[i][j]*gtable[col[i][j]]/150;
	if (c>255) c=255;
	putc(c,outfile);
	c = shades[i][j]*btable[col[i][j]]/150;
	if (c>255) c=255;
	putc(c,outfile);
      }
  } else {
    for (j=0; j<Height; j++)
      for (i=0; i<Width; i++) {
	putc(rtable[col[i][j]],outfile);
	putc(gtable[col[i][j]],outfile);
	putc(btable[col[i][j]],outfile);
      }
  }
  fclose(outfile);
}
 
void printheights(outfile) /* prints heightfield */
FILE *outfile;
{
  int i,j;

  for (j=0; j<Height; j++) {
    for (i=0; i<Width; i++)
      fprintf(outfile,"%d ",heights[i][j]);
    putc('\n',outfile);
  }
  fclose(outfile);
}
      
double log2(x)
double x;
{ return(log(x)/log(2.0)); }

void print_error()
{
  fprintf(stderr,"Usage: planet [options]\n\n");
  fprintf(stderr,"options:\n");
  fprintf(stderr,"  -s seed           Specifies seed as number between 0.0 and 1.0\n");
  fprintf(stderr,"  -w width          Specifies width in pixels, default = 800\n");
  fprintf(stderr,"  -h height         Specifies height in pixels, default = 600\n");
  fprintf(stderr,"  -m magnification  Specifies magnification, default = 1.0\n");
  fprintf(stderr,"  -o output-file    Specifies output file, default = stdout\n");
  fprintf(stderr,"  -l longitude      Specifies longitude of centre in degrees, default = 0.0\n");
  fprintf(stderr,"  -L latitude       Specifies latitude of centre in degrees, default = 0.0\n");
  fprintf(stderr,"  -g gridsize       Specifies vertical gridsize in degrees, default = 0.0 (no grid)\n");
  fprintf(stderr,"  -G gridsize       Specifies horisontal gridsize in degrees, default = 0.0 (no grid)\n");
  fprintf(stderr,"  -c                Colour depends on latitude (default: only altitude)\n");
  fprintf(stderr,"  -C                Use lighter colours\n");
  fprintf(stderr,"  -N nocols         Use nocols colours (default = 256)\n");
  fprintf(stderr,"  -S smooth         Amount of smoothing (default = 0.0)  \n");
  fprintf(stderr,"  -B                Use ``bumpmap'' shading\n");
  fprintf(stderr,"  -b                Reverse background colour (default: black)\n");
  fprintf(stderr,"  -i init_alt       Specifies initial altitude (default = -0.015)\n");
  fprintf(stderr,"  -V                Initial altitude variation (default = 0.13)\n");
  fprintf(stderr,"  -v                Change in altitude variation (default = 0.9)\n");
  fprintf(stderr,"  -pprojection      Specifies projection: m = Mercator (default)\n");
  fprintf(stderr,"                                          p = Peters\n");      fprintf(stderr,"                                          q = Square\n");
  fprintf(stderr,"                                          s = Stereographic\n");
  fprintf(stderr,"                                          o = Orthographic\n");
  fprintf(stderr,"                                          g = Gnomonic\n");
  fprintf(stderr,"                                          a = Area preserving azimuthal\n");
  fprintf(stderr,"                                          c = Conical (conformal)\n");
  fprintf(stderr,"                                          M = Mollweide\n");
  fprintf(stderr,"                                          S = Sinusoid\n");
  fprintf(stderr,"                                          h = heightfield\n");
  fprintf(stderr,"                                          f = Find match, see manual\n");
  exit(0);
}

/* With the -pf option a map must be given on standard input.  */
/* This map is 11 lines of 24 characters. The characters are:  */
/*    . : very strong preference for water (value=8)	       */
/*    , : strong preference for water (value=4)		       */
/*    : : preference for water (value=2)		       */
/*    ; : weak preference for water (value=1)		       */
/*    - : don't care (value=0)				       */
/*    * : weak preference for land (value=1)		       */
/*    o : preference for land (value=2)			       */
/*    O : strong preference for land (value=4)		       */
/*    @ : very strong preference for land (value=8)	       */
/*                                                             */
/* Each point on the map corresponds to a point on a 15 grid. */
/*							       */
/* The program tries seeds starting from the specified and     */
/* successively outputs the seed (and rotation) of the best    */
/* current match, together with a small map of this.	       */
/* This is all ascii, no bitmap is produced.		       a*/

